from abc import ABC, abstractmethod
from datetime import datetime
from logging import Logger
from pathlib import Path
from typing import Optional, Union, TYPE_CHECKING, Literal, Type, Iterable, List, Any

from mcdreforged.api.rtext import RTextBase, RColor
from mcdreforged.api.types import CommandSource
from pydantic import BaseModel, ValidationError, TypeAdapter
from ruamel.yaml import YAML

from calyx_lib import constants
from calyx_lib.config.pydantic import PydanticValidationErrorMessage, \
    ConfigSerializationContext
from calyx_lib.config.yaml import any_to_yaml, ConfigComment, \
    CommentContext, CommentCarrier, adjust_comment_indentation
from calyx_lib.translation.text import RTextBlossomTranslation
from calyx_lib.translation.translator import BlossomTranslator
from calyx_lib.typing import MessageText, PathStr, Subscriptable, TranslationKeyDictRich
from calyx_lib.utils import touch_directory

if TYPE_CHECKING:
    from mcdreforged.api.rtext import RTextMCDRTranslation
    from typing import Dict


class BlossomBaseInterface(ABC):
    class __ConfigProcessLoggingHandler:
        def __init__(
                self,
                base: "BlossomBaseInterface",
                echo_in_console: bool,
                source_to_reply: Optional[CommandSource]
        ):
            self.base = base
            self.echo_in_console = echo_in_console
            self.source_to_reply: Optional[CommandSource]
            if isinstance(source_to_reply, CommandSource):
                self.source_to_reply = source_to_reply
            else:
                self.source_to_reply = None

        def info(self, msg: MessageText):
            if self.echo_in_console:
                self.base.logger.info(msg)
            if self.source_to_reply is not None:
                self.source_to_reply.reply(msg)

        def warning(self, msg: MessageText):
            if self.echo_in_console:
                self.base.logger.warning(msg)
            if self.source_to_reply is not None:
                msg = RTextBase.from_any(msg).set_color(RColor.yellow)
                self.source_to_reply.reply(msg)

    def __init__(self):
        self.logger = self.get_logger()
        self.translator = BlossomTranslator(self)
        for lang in ["zh_cn", "en_us"]:
            target_file_path = Path(constants.SELF_PACKAGE_PATH) / "lang" / f"{lang}.yml"
            if target_file_path.is_file():
                self.translator.register_translation_file(
                    target_file_path, encoding="utf8"
                )

    @abstractmethod
    def get_logger(self) -> Logger:
        ...

    @abstractmethod
    def get_language(self):
        ...

    def dtr(self, translation_dict: TranslationKeyDictRich, *args, **kwargs):
        def fake_tr(
            translation_key: str,
            *inner_args,
            language: Optional[str] = None,
            _mcdr_tr_language: Optional[str] = None,
            _mcdr_tr_allow_failure: bool = True,
            _calyx_log_error_message: bool = True,
            _calyx_default_fallback: str = "<Translation failed>",
            **inner_kwargs,
        ) -> MessageText:
            if language is not None and _mcdr_tr_language is None:
                _mcdr_tr_language = language
            try:
                return self.translator.translate_from_dict(
                    translation_dict,
                    *inner_args,
                    _mcdr_tr_language=_mcdr_tr_language,
                    **inner_kwargs,
                )
            except Exception as e:
                lang_text = ", ".join(
                    [f'"{lang}"' for lang in self.translator.fallback_language_order]
                )
                error_message = (
                    f"Error translate text from dict to language {lang_text}: {str(e)}"
                )
                if _mcdr_tr_allow_failure:
                    if _calyx_log_error_message:
                        self.logger.error(error_message)
                    return _calyx_default_fallback
                else:
                    raise e

        return RTextBlossomTranslation(
            "", *args, **kwargs
        ).set_translator(fake_tr).set_language_getter(self.get_language)

    def tr(
            self,
            translation_key: str,
            *args,
            language: Optional[str] = None,
            _mcdr_tr_language: Optional[str] = None,
            _mcdr_tr_allow_failure: bool = True,
            _calyx_default_fallback: Optional[MessageText] = None,
            _calyx_log_error_message: bool = True,
            **kwargs
    ) -> MessageText:
        target_lang = _mcdr_tr_language or language or self.get_language()
        return self.translator.translate(
            translation_key,
            *args,
            _mcdr_tr_language=target_lang,
            _mcdr_tr_allow_failure=_mcdr_tr_allow_failure,
            _calyx_default_fallback=_calyx_default_fallback,
            _calyx_log_error_message=_calyx_log_error_message,
            **kwargs
        )

    def rtr(
            self,
            translation_key: str,
            *args,
            _mcdr_tr_allow_failure: bool = True,
            _calyx_default_fallback: Optional[MessageText] = None,
            _calyx_log_error_message: bool = True,
            **kwargs
    ) -> Union["RTextMCDRTranslation", "RTextBlossomTranslation"]:
        return RTextBlossomTranslation(
            translation_key,
            *args,
            _mcdr_tr_allow_failure=_mcdr_tr_allow_failure,
            _calyx_default_fallback=_calyx_default_fallback,
            _calyx_log_error_message=_calyx_log_error_message,
            **kwargs,
        ).set_language_getter(self.get_language).set_translator(self.tr)

    def __get_default_comment_context(self):
        def tr(
                translation_key: str,
                *args, **kwargs
        ):
            return self.tr(
                translation_key,
                *args,
                _calyx_default_fallback=translation_key,
                _calyx_log_error_message=False,
                **kwargs
            )
        headline = ConfigComment(
            text=self.tr(
                'config.saving.comments.saving_at',
                datetime.now().strftime(
                    str(self.tr('general.format.datetime'))
                )
            ),
            priority=float('-inf'),
            ignore_global_wrapper=False,
        )

        return CommentContext(
            key_comments={}, global_wrapper=tr, headlines=[headline], eof=[]
        )

    def save_config(
            self,
            file_name: PathStr,
            config: BaseModel,
            echo_in_console: bool = True,
            source_to_reply: Optional[CommandSource] = None,
            failure_policy: "Literal['regen', 'raise']" = 'regen',
            encoding: str = 'utf8',
            should_generate_comment: bool = True,
            optional_context: Optional[CommentContext] = None
    ):
        log_handler = self.__ConfigProcessLoggingHandler(
            self, echo_in_console=echo_in_console, source_to_reply=source_to_reply
        )

        file_path = Path(file_name)
        if file_path.is_dir():
            file_path.rmdir()

        context = self.__get_default_comment_context()
        if should_generate_comment:
            context = optional_context or context
        try:
            serialized = config.model_dump(
                exclude_none=True,
                context=ConfigSerializationContext(
                    global_wrapper=context.global_wrapper
                )
            )
        except Exception as exc:
            if failure_policy == 'raise':
                raise exc
            serialized = config.__class__().model_dump(
                exclude_none=True,
                context=ConfigSerializationContext()
            )

        should_generate_comment = should_generate_comment and isinstance(
            serialized, CommentCarrier
        )
        if should_generate_comment:
            context.apply_comment_to_carrier(serialized)
        with open(file_path, mode='w', encoding=encoding) as f:
            if should_generate_comment:
                context.apply_comment_to_stream(context.headlines, f)
            f.write(adjust_comment_indentation(any_to_yaml(serialized)))
            if should_generate_comment:
                context.apply_comment_to_stream(context.eof, f)
        log_handler.info(
            self.rtr(
                'config.saving.config_saved', file=str(file_path)
            )
        )

    def load_config(
        self,
        file_path: PathStr,
        model_class: Type[BaseModel],
        echo_in_console: bool = True,
        source_to_reply: Optional[CommandSource] = None,
        encoding: str = "utf8",
        failure_policy: Literal['regen', 'raise'] = "regen",
        should_generate_comment: bool = True,
    ):
        log_handler = self.__ConfigProcessLoggingHandler(
            self, echo_in_console=echo_in_console, source_to_reply=source_to_reply
        )
        file_path = Path(file_path)
        requires_save = False

        touch_directory(file_path.parent)
        if file_path.is_dir():
            file_path.rmdir()
        if not file_path.is_file():
            cfg_final = model_class()
            self.save_config(
                file_path, cfg_final,
                echo_in_console=echo_in_console,
                source_to_reply=source_to_reply,
                encoding=encoding,
                failure_policy=failure_policy,
                should_generate_comment=should_generate_comment
            )
            log_handler.warning(
                self.rtr('config.loading.file_not_found', file=str(file_path))
            )
            return cfg_final

        default = model_class()
        default_dict_included_none = default.model_dump()
        default_dict_excluded_none = default.model_dump(exclude_none=True)

        comment_context = None
        if should_generate_comment:
            comment_context = self.__get_default_comment_context()

        try:
            with open(file_path, 'r', encoding=encoding) as f:
                raw_data = YAML(typ='safe').load(f)
            for k, v in default_dict_excluded_none.items():
                if k not in raw_data.keys():
                    raw_data[k] = v
                    log_handler.warning(
                        self.rtr('config.loading.item_lost', key=k)
                    )
                    if comment_context is not None:
                        comment_context.add_comment(
                            (k, ),
                            ConfigComment(
                                self.tr('config.saving.comments.fixed_missing', key=k),
                                priority=float('-inf'),
                                ignore_global_wrapper=True
                            )
                        )
                    requires_save = True

            try:
                cfg_final = model_class.model_validate(raw_data)
            except ValidationError as exc:
                requires_save = True
                exc: ValidationError    # type: ignore
                # Yeet both pycharm and mypy warnings :<
                errors = TypeAdapter(
                    List[PydanticValidationErrorMessage]
                ).validate_python(
                    exc.errors()
                )

                fixed = {}  # type: Dict[tuple, List[PydanticValidationErrorMessage]]

                def fix_nested_values(
                        error: PydanticValidationErrorMessage,
                        item: Subscriptable,
                        default: Subscriptable,
                        remaining: Iterable[Any],
                        consumed: Optional[List[Any]] = None,
                ) -> bool:
                    current_path = list(remaining)

                    if not current_path:
                        return True
                    current_index = current_path.pop(0)
                    consumed = consumed or []
                    try:
                        if isinstance(default, list):
                            current_default = default[0]
                        else:
                            current_default = default[current_index]
                    except:
                        return True
                    consumed.append(current_index)
                    try:
                        current_value = item[current_index]
                    except:
                        item[current_index] = current_default
                        return False
                    fix_this_layer = fix_nested_values(
                        error, current_value, current_default, current_path, consumed
                    )
                    path_tuple = tuple(consumed)
                    if fix_this_layer:
                        prev = item[current_index]
                        if path_tuple not in fixed:
                            item[current_index] = current_default
                        if fixed.get(path_tuple) is None:
                            fixed[path_tuple] = []
                        fixed[path_tuple].append(
                            error.model_copy(update={'input': prev})
                        )
                    return False

                self.logger.debug("  -- Pydantic validation errors -- ")
                for e in errors:
                    self.logger.debug(f'Error found at {".".join(str(i) for i in e.loc)}')
                    self.logger.debug(f'  Error message: {e.msg}')
                    self.logger.debug(f'  Type: {e.type}')
                    self.logger.debug(f'  Input value: {e.input}')
                    requires_fix = fix_nested_values(
                        e, raw_data, default_dict_included_none, e.loc
                    )
                    if requires_fix:
                        self.logger.debug(">> Error can't be fixed, regenerating... <<")
                        raise

                if comment_context is not None:
                    for k, error_list in fixed.items():
                        error_text = []
                        for e in error_list:
                            error_text.append('- ' + e.msg)
                        input_value = any_to_yaml(error_list[0].input)
                        if len(input_value.splitlines()) > 1:
                            input_value = '\n' + input_value
                        comment = ConfigComment(
                            self.rtr(
                                'config.saving.comments.fixed_type_error',
                                key='.'.join([str(char) for char in k]),
                                errors='\n'.join(error_text),
                                value=input_value
                            ),
                            priority=float('-inf')
                        )
                        comment_context.add_comment(k, comment)

                for pt in fixed:
                    log_handler.warning(
                        self.rtr(
                            "config.loading.type_fixed", key='.'.join(
                                [str(i) for i in pt]
                            )
                        )
                    )
                cfg_final = model_class.model_validate(raw_data)
        except Exception as exc:
            if failure_policy == 'raise':
                raise
            requires_save = True
            cfg_final = model_class()
            log_handler.warning(self.rtr('config.loading.yaml_syntax_error'))
            self.logger.debug(f"Error reason: {exc}")
        if requires_save:
            self.save_config(
                file_path,
                cfg_final,
                echo_in_console=echo_in_console,
                source_to_reply=source_to_reply,
                encoding=encoding,
                failure_policy=failure_policy,
                should_generate_comment=should_generate_comment,
                optional_context=comment_context
            )
        log_handler.info(self.rtr('config.loading.config_loaded'))

        return cfg_final



from calyx_lib.config.yaml import ConfigComment, CommentContext
from calyx_lib.config.pydantic import CommentedModel, BLANK, Blankable


__all__ = [
    "ConfigComment",
    "CommentContext",
    "CommentedModel",
    "BLANK",
    "Blankable"
]



import threading

from contextlib import contextmanager
from typing import Callable, List, Optional, Union, Iterable
from typing_extensions import Self, override, Unpack

from mcdreforged.api.rtext import RTextBase, RStyle, RColor, RAction, RTextMCDRTranslation

from calyx_lib import utils
from calyx_lib.typing import TranslateFunc


class RTextBlossomTranslation(RTextBase):
    """
    The alternative translation text component out of MCDR for plugins

    When MCDR is running, it will use the :meth:`~mcdreforged.plugin.si.server_interface.ServerInterface.tr` method
    in :class:`~mcdreforged.plugin.si.server_interface.ServerInterface` class as the translating method,
    and the language of MCDR as the fallback translation language

    .. versionadded:: v2.1.0
    """

    __TLS = threading.local()
    __TLS.language = None

    def __init__(self, translation_key: str, *args, **kwargs):
        """
        Create a :class:`RTextBlossomTranslation` component
        with necessary parameters for translation

        :param translation_key: The translation key
        :param args: The translation arguments
        :param kwargs: The translation keyword arguments
        """
        def default_tr_func(key: str, *va, **kw) -> str:
            return str(key)

        self.translation_key: str = translation_key
        self.args = args
        self.kwargs = kwargs
        self.__tr_func: TranslateFunc = default_tr_func
        self.__post_process: List[Callable[[RTextBase], RTextBase]] = []
        self.__language_getter: Callable[[], str] = lambda: 'en_us'

    def set_translator(self, translate_function: TranslateFunc) -> 'Self':
        self.__tr_func = translate_function
        return self

    def set_language_getter(self, language_getter: Callable[[], str]):
        self.__language_getter = language_getter
        return self

    def get_language(self):
        return self.__language_getter()

    def __get_translated_text(self) -> RTextBase:
        language = getattr(self.__TLS, 'language', None)
        if language is None:
            language = self.get_language()
        processed_text = self.__tr_func(
            self.translation_key,
            *self.args,
            **self.kwargs,
            _mcdr_tr_language=language
        )
        processed_text = RTextBase.from_any(processed_text)
        for process in self.__post_process:
            processed_text = process(processed_text)
        return processed_text

    @classmethod
    @contextmanager
    def language_context(cls, language: str):
        """
        Create a context where all :class:`RTextBlossomTranslation`
        will use the given language to translate within

        It's mostly used when you want a translated str or Minecraft json text object
        corresponding to this component under a specific language

        MCDR will automatically apply this context with :ref:`user's
        preferred language <preference-language>`
        right before sending messages to a player or the console

        Example::

            def log_message_line_by_line(server: ServerInterface):
                with RTextMCDRTranslation.language_context('en_us'):
                    text: RTextMCDRTranslation = server.rtr('my_plugin.some_message')
                    # The translation operation happens here
                    text_as_str: str = text.to_plain_text()
                    server.logger.info('Lines of my translation')
                    for line in text_as_str.splitlines():
                        server.logger.info('- {}'.format(line))


        :param language: The language to be used during translation inside the context
        """
        prev = getattr(cls.__TLS, 'language', None)
        cls.__TLS.language = language
        try:
            yield
        finally:
            cls.__TLS.language = prev

    @override
    def to_json_object(self, **kwargs: Unpack[RTextBase.ToJsonKwargs]) -> Union[dict, list]:
        return self.__get_translated_text().to_json_object(**kwargs)

    @override
    def to_plain_text(self) -> str:
        return self.__get_translated_text().to_plain_text()

    @override
    def to_colored_text(self) -> str:
        return self.__get_translated_text().to_colored_text()

    @override
    def to_legacy_text(self) -> str:
        return self.__get_translated_text().to_legacy_text()

    @override
    def copy(self) -> 'Self':
        copied = self.__class__(self.translation_key, *self.args, **self.kwargs)
        copied.__tr_func = self.__tr_func
        copied.__post_process = self.__post_process.copy()
        return copied

    @override
    def set_color(self, color: RColor) -> Self:
        def add_color(rt: RTextBase):
            return rt.set_color(color)
        self.__post_process.append(add_color)
        return self

    @override
    def set_styles(self, styles: Union[RStyle, Iterable[RStyle]]) -> Self:
        def set_styles(rt: RTextBase):
            return rt.set_styles(styles)
        self.__post_process.append(set_styles)
        return self

    @override
    def set_click_event(self, action: RAction, value: str) -> Self:
        def set_click_event(rt: RTextBase):
            return rt.set_click_event(action, value)
        self.__post_process.append(set_click_event)
        return self

    @override
    def set_hover_text(self, *args) -> Self:
        def set_hover_text(rt: RTextBase):
            return rt.set_hover_text(*args)
        self.__post_process.append(set_hover_text)
        return self

    def __repr__(self) -> str:
        return utils.represent(self)



import os
import re
from pathlib import Path
from typing import Optional, Collection, Any
from typing import Union
from zipfile import ZipFile

from calyx_lib.typing import PathStr

__all__ = [
    'touch_directory',
    'clean_console_color_code',
    'clean_minecraft_color_code',
    'capitalize',
    'to_camel_case',
    'list_bundled_file',
    'represent'
]

"""
touch_directory(), clean_minecraft_color_code(), clean_console_color_code(),
represent(), list_bundled_file
Copied from MCDReforged(https://mcdreforged.com) v2.14.7
Licensed under LGPL v3.0 (only)
"""


def touch_directory(directory_path: Union[Path, str]) -> None:
    if not os.path.isdir(directory_path):
        os.makedirs(directory_path, exist_ok=True)


def clean_minecraft_color_code(text: str):
    return re.compile(r'§[a-z0-9]').sub('', str(text))


def clean_console_color_code(text: str):
    return re.compile(r'\033\[(\d+(;\d+)*)?m').sub('', text)


def capitalize(string: str) -> str:
    char_list = list(string)
    if len(char_list) > 0:
        char_list[0] = char_list[0].upper()
    return "".join(char_list)


def to_camel_case(string: str, divider: str = " ", upper: bool = True) -> str:
    word_list = [capitalize(item) for item in string.split(divider)]
    if not upper:
        first_word_char_list = list(word_list[0])
        first_word_char_list[0] = first_word_char_list[0].lower()
        word_list[0] = "".join(first_word_char_list)
    return "".join(word_list)


def list_bundled_file(package_path: PathStr, directory_name: PathStr):
    if os.path.isdir(package_path):
        return os.listdir(os.path.join(package_path, directory_name))
    with ZipFile(package_path, 'r') as zip_file:
        result = []
        directory_name = str(directory_name).replace('\\', '/').rstrip('/\\') + '/'
        for file_info in zip_file.infolist():
            # is inside the dir and is directly inside
            if file_info.filename.startswith(directory_name):
                file_name = file_info.filename.replace(directory_name, '', 1)
                if len(file_name) > 0 and '/' not in file_name.rstrip('/'):
                    result.append(file_name)
    return result


def represent(
        obj: Any,
        fields: Optional[dict] = None,
        *,
        blacklist: Collection[str] = (),
        parentheses: str = '()'
) -> str:
    """
    aka repr
    """
    if fields is None:
        fields = {k: v for k, v in vars(obj).items() if not k.startswith('_')}
    blacklist = set(blacklist)
    return ''.join([
        type(obj).__name__,
        parentheses[0],
        ', '.join([
            f'{k}={v!r}'
            for k, v in fields.items()
            if k not in blacklist
        ]),
        parentheses[1],
    ])



import os

NAME = 'CalyxLib'
PACKAGE_NAME = 'calyx_lib'

VERSION = '0.1.0-alpha.2'
VERSION_PYPI = '0.1.0a2'

GITHUB_URL = 'https://github.com/Lazy-Bing-Server/CalyxLib-Python'

SELF_PACKAGE_PATH = os.path.abspath(os.path.dirname(__file__))
LANGUAGE_FILE_SUFFIX = '.yml'
DEFAULT_LANGUAGE = 'en_us'

SELF_LANG_FOLDER = 'resources/lang'
DESCRIPTION = ("Alternative MCDReforged API for MCDR plugins that "
               "requires running in other environments")




from pathlib import Path
from typing import Union, Callable, Any, Dict, Protocol

from mcdreforged.api.rtext import RTextBase
from mypy_extensions import VarArg, KwArg

PathStr = Union[str, Path]

MessageText = Union[str, RTextBase]
TranslateFunc = Callable[
    [str, VarArg(Any), KwArg(Any)], MessageText
]

# language -> text
TranslationLanguageDict = Dict[str, str]
# key -> text
TranslationKeyDict = Dict[str, str]
# key -> text/(key -> text/(...))
TranslationKeyDictNested = Dict[str, Union[str, 'TranslationKeyDictNested']]
# language -> text
TranslationKeyDictRich = Dict[str, MessageText]
# key -> (language -> text)
TranslationStorage = Dict[str, TranslationLanguageDict]


class Subscriptable(Protocol):
    def __getitem__(self, item) -> Any:
        pass

    def __setitem__(self, item, value) -> Any:
        pass


CommentTextWrapper = Callable[[MessageText], MessageText]



import contextlib
import json
import os
from logging import Logger
from pathlib import Path
from threading import RLock
from typing import Optional, List, Dict, Union, Callable, Any, TYPE_CHECKING
from zipfile import ZipFile

from mcdreforged.api.rtext import RTextBase, RTextMCDRTranslation
from mcdreforged.api.types import ServerInterface
from ruamel.yaml import YAML

from calyx_lib.utils import list_bundled_file
from calyx_lib.typing import (
    PathStr,
    MessageText,
    TranslationKeyDictNested,
    TranslationKeyDict,
    TranslationLanguageDict,
    TranslationStorage
)

if TYPE_CHECKING:
    from calyx_lib.interface.blossom_base_interface import BlossomBaseInterface

_NONE = object()


__all__ = [
    "BlossomTranslator"
]


class BlossomTranslator:
    yaml = YAML(typ='safe')

    def __init__(self, base_interface: "BlossomBaseInterface"):
        self.__base_interface = base_interface
        self.__storage: TranslationStorage = {}
        self.__lock = RLock()
        self.__fallback_language_order = ['en_us']

        psi = ServerInterface.psi_opt()
        if psi is not None:
            self.set_language(psi.get_mcdr_language())

    @property
    def logger(self) -> Logger:
        return self.__base_interface.logger

    @property
    def available(self) -> bool:
        return len(self.__storage) != 0

    @property
    def fallback_language_order(self):
        return self.__fallback_language_order

    def set_language(self, language: str):
        with self.__lock:
            if language in self.__fallback_language_order:
                self.__fallback_language_order.remove(language)
            self.__fallback_language_order = [language] + self.__fallback_language_order

    def remove_language(self, language: str):
        with self.__lock:
            if language not in self.__fallback_language_order:
                return
            self.__fallback_language_order.remove(language)
            if language == 'en_us':
                self.__fallback_language_order.append(language)

    def register_translation(
            self, translation_dict: TranslationKeyDictNested, language: str
    ):
        def get_full_key_value_map(
                target_dict: TranslationKeyDictNested,
                result_dict: Optional[TranslationKeyDict] = None,
                current_layer: Optional[List[str]] = None
        ):
            if current_layer is None:
                current_layer = []
            if result_dict is None:
                result_dict = {}
            for k, v in target_dict.items():
                this_layer = current_layer.copy()
                this_layer.append(k)
                if isinstance(v, dict):
                    get_full_key_value_map(
                        v, result_dict=result_dict, current_layer=this_layer
                    )
                else:
                    result_dict['.'.join(this_layer)] = str(v)
            return result_dict

        translation_dict = get_full_key_value_map(translation_dict)
        for key, value in translation_dict.items():
            if (
                    key not in self.__storage.keys()
                    or
                    not isinstance(self.__storage[key], dict)
            ):
                self.__storage[key] = {}
            self.__storage[key][language] = value  # type: ignore

    def register_translation_file(
            self,
            file_path: PathStr,
            bundle_path: Optional[PathStr] = None,
            error_handler: Optional[Callable[[Exception], Any]] = None,
            encoding: str = 'utf8'
    ) -> bool:
        file_name = os.path.basename(file_path)
        if '.' not in list(file_name):
            return False
        language, file_extension = file_name.rsplit('.', 1)
        if file_extension in ['json', 'yml', 'yaml']:
            try:
                if bundle_path is not None:
                    with ZipFile(bundle_path).open(str(file_path)) as file:
                        text = file.read().decode(encoding=encoding)
                else:
                    with open(file_path, 'r', encoding=encoding) as f:
                        text = f.read()
                if file_extension == 'json':
                    translation_dict = json.loads(text)
                else:
                    translation_dict = self.yaml.load(text)
                if isinstance(translation_dict, dict):
                    self.register_translation(translation_dict, language)
                return True
            except Exception as e:
                if error_handler is not None:
                    error_handler(e)
        return False

    def register_bundled_translations(self, package: PathStr, target_folder: PathStr):
        package_path = Path(package)
        target_folder_path = Path(target_folder)
        for file_name in list_bundled_file(package_path, target_folder_path):
            if package_path.is_dir():
                translation_file_path = package_path / target_folder_path / file_name
                registered = self.register_translation_file(translation_file_path)
            else:
                registered = self.register_translation_file(
                    target_folder_path / file_name, package_path
                )
            if not registered:
                self.logger.debug(
                    'Skipping unknown translation file {} in {}'.format(
                        file_name, repr(self))
                )

    def has_translation(
            self, translation_key: str, override_language: Optional[str] = None
    ):
        trans = self.__storage.get(translation_key)
        if isinstance(trans, dict):
            with self.language_context(override_language):
                for lang in self.__fallback_language_order:
                    if lang in trans.keys():
                        return True
        return False

    @contextlib.contextmanager
    def language_context(self, language: Optional[str]):
        with self.__lock:
            language_order = self.__fallback_language_order
            self.__fallback_language_order = self.__fallback_language_order.copy()
            try:
                if language is not None:
                    self.set_language(language)
                yield
            finally:
                self.__fallback_language_order = language_order

    def translate_from_dict(self, translation_dict: TranslationLanguageDict, *args, **kwargs):
        translated_raw_text = _NONE
        for lang in self.__fallback_language_order:
            translated_raw_text = translation_dict.get(lang, _NONE)
            if translated_raw_text is not _NONE:
                break
        # Allow null value in translation files
        if translated_raw_text is None:
            translated_formatter = ''
        elif isinstance(translated_raw_text, str):
            translated_formatter = translated_raw_text
        elif translated_raw_text is _NONE:
            raise KeyError("Translation key does not exist")
        else:
            raise TypeError("Unexpected type found in translation storage")

        use_rtext = any(
            [isinstance(e, RTextBase) for e in list(args) + list(kwargs.values())]
        )
        try:
            if use_rtext:
                return RTextBase.format(translated_formatter, *args, **kwargs)
            else:
                return translated_formatter.format(*args, **kwargs)
        except Exception as e:
            raise ValueError(f'Failed to apply args {args} and kwargs {kwargs} '
                             f'to translated_text {translated_raw_text}: {str(e)}')

    def translate(
            self,
            translation_key: str,
            *args,
            language: Optional[str] = None,
            _mcdr_tr_language: Optional[str] = None,
            _mcdr_tr_allow_failure: bool = True,
            _calyx_default_fallback: Optional[MessageText] = None,
            _calyx_log_error_message: bool = True,
            **kwargs
    ) -> MessageText:
        if not self.available:
            raise RuntimeError('Illegal translate request before translation loading')
        _mcdr_tr_language = _mcdr_tr_language or language
        _calyx_default_fallback = _calyx_default_fallback or translation_key
        translation_dict = self.__storage.get(translation_key, {})

        with self.language_context(language=_mcdr_tr_language):
            try:
                return self.translate_from_dict(translation_dict, *args, **kwargs)
            except Exception as e:
                lang_text = ', '.join(
                    [f'"{lang}"' for lang in self.__fallback_language_order]
                )
                error_message = 'Error translate text "{}" to language {}: {}'.format(
                    translation_key, lang_text, str(e)
                )
                if _mcdr_tr_allow_failure:
                    if _calyx_log_error_message:
                        self.logger.error(error_message)
                    return _calyx_default_fallback
                else:
                    raise e

    def dict_tr(self, translation_dict: Dict[str, str], *args, **kwargs):
        pass



import copy
import dataclasses
from io import StringIO
from typing import Any, Literal, Union, List, Dict, Iterable, Optional, IO

from mcdreforged.api.rtext import RTextBase
from ruamel.yaml import YAML, CommentedMap, RoundTripRepresenter

from calyx_lib.typing import MessageText, CommentTextWrapper, Subscriptable


class ConfigComment:
    def __init__(
        self,
        text: MessageText,
        loc: Literal['before', 'eol'] = 'before',
        wrapper: CommentTextWrapper = lambda text: text,
        priority: Union[int, float] = 1000,
        ignore_global_wrapper: bool = False
    ):
        self.text = text
        self.loc = loc
        self.wrapper = wrapper
        self.priority = priority
        self.ignore_global_wrapper = ignore_global_wrapper

    def get_text(self):
        return self.wrapper(self.text)

    @property
    def is_before(self):
        return self.loc == 'before'

    @property
    def is_eol(self):
        return self.loc == "eol"

    def copy(self):
        return copy.copy(self)


@dataclasses.dataclass
class CommentContext:
    key_comments: Dict[tuple, List[ConfigComment]]
    global_wrapper: CommentTextWrapper
    headlines: List[ConfigComment]
    eof: List[ConfigComment]

    def add_comment(self, key: tuple, comment: ConfigComment):
        if key not in self.key_comments.keys():
            self.key_comments[key] = []
        self.key_comments[key].append(comment)

    def __get_comment_text(self, comment: ConfigComment):
        text = comment.get_text()
        if not comment.ignore_global_wrapper:
            text = self.global_wrapper(text)
        return RTextBase.from_any(text).to_plain_text()

    def apply_comment_to_stream(
            self, comments: List[ConfigComment], stream: IO, indent: int = 0
    ):
        raw_text_lines = []
        for comment in comments:
            raw_text_lines += list(self.__get_comment_text(comment).splitlines())

        line_prefix = indent * ' ' + "# "
        comment_lines = []
        for line in raw_text_lines:
            if str(line).strip() != '':
                comment_lines.append(line_prefix + line)
            else:
                comment_lines.append(line)
        if comment_lines:
            stream.write('\n'.join(comment_lines) + '\n\n')

    def apply_comment_to_carrier(self, carrier: "CommentCarrier"):
        for key, comments in self.key_comments.items():
            for comment in comments:
                carrier.set_comment_to_nested_key(key, comment)


class CommentCarrier:
    def __init__(self, data: dict):
        self.data = data
        self.comments = {}  # type: Dict[str, List[ConfigComment]]
        self.global_wrapper = lambda text: text

    def get_comment_text(self, comment: ConfigComment):
        text = comment.get_text()
        if not comment.ignore_global_wrapper:
            text = self.global_wrapper(text)
        return RTextBase.from_any(text).to_plain_text()

    def to_commented_map(self):
        data = CommentedMap(self.data)

        for key, comments in self.comments.items():
            before_comments = []
            for cmt in sorted(comments, key=lambda item: item.priority):
                if cmt.is_before:
                    before_comments.append(self.get_comment_text(cmt))
                elif cmt.is_eol:
                    data.yaml_add_eol_comment(
                        key=key, comment=self.get_comment_text(cmt)
                    )
            if before_comments:
                before_comments.reverse()
                before_comment_text = '\n'.join(before_comments)
                data.yaml_set_comment_before_after_key(
                    key=key, before=before_comment_text
                )
        return data

    def set_comment_to_nested_key(
            self, key: Iterable[str], comment: ConfigComment
    ):
        key_list = list(key)
        data: Subscriptable = self.data
        comment = comment.copy()
        while True:
            current_key = key_list.pop(0)
            try:
                current_value = data[current_key]
                if not key_list:
                    if current_key not in self.comments.keys():
                        self.comments[current_key] = []
                    self.comments[current_key].append(comment)
                    break
                elif isinstance(current_value, CommentCarrier):
                    current_value.set_comment_to_nested_key(
                        key_list, comment
                    )
                    break
                else:
                    data = current_value
            except:
                break

    @staticmethod
    def represent(representer: RoundTripRepresenter, item: "CommentCarrier"):
        return representer.represent_dict(item.to_commented_map())

    def __iter__(self):
        for item in self.data.items():
            yield item


def get_yaml():
    yaml = YAML(typ='rt')
    yaml.width = 1048576
    yaml.allow_unicode = True
    yaml.representer.add_representer(CommentCarrier, CommentCarrier.represent)
    return yaml


def any_to_yaml(any_item: Any):
    yaml = get_yaml()
    with StringIO() as stream:
        yaml.dump(any_item, stream)
        stream.seek(0)
        return stream.read().strip()


# Thanks to gemini
def adjust_comment_indentation(yaml_string: str):
    lines = list(yaml_string.strip().splitlines())
    lines.reverse()
    adjusted_lines = []
    indentation_level = 0
    indentation_char = ' '  # Assuming spaces for indentation

    for i, line in enumerate(lines):
        stripped_line = line.lstrip()
        current_indentation = len(line) - len(stripped_line)

        if stripped_line.startswith('#'):
            # It's a comment
            if i > 0:
                # Get the indentation of the previous non-comment line
                prev_indentation = indentation_level * len(indentation_char)
                adjusted_lines.append(
                    f"{indentation_char * prev_indentation}{stripped_line}"
                )
            else:
                # First line comment, keep its original indentation
                adjusted_lines.append(line)
        else:
            # It's a non-comment line, update the indentation level
            indentation_level = current_indentation // len(indentation_char)
            adjusted_lines.append(line)

    adjusted_lines.reverse()
    return '\n'.join(adjusted_lines)



import re
from contextlib import contextmanager
from logging import Logger
from queue import Queue
from typing import Union, Optional, List

from mcdreforged.api.event import MCDRPluginEvents
from mcdreforged.api.types import PluginServerInterface, Info


__all__ = [
    'CommandQueries'
]


class QueryItem:
    def __init__(self, command: str, patterns: List[Union[str, re.Pattern]]):
        self.command = command
        self.patterns = [re.compile(pattern) for pattern in patterns]
        self.queue: Queue[re.Match] = Queue()

    def get(self, timeout: float = 3.0) -> Optional[re.Match]:
        return self.queue.get(block=True, timeout=timeout)


class CommandQueries:
    def __init__(
            self,
            server: "PluginServerInterface",
            logger: Optional[Logger] = None,
            allow_use_same_line: bool = True
    ):
        self.__server = server
        self.__logger = logger or server.logger
        self.__items: List[QueryItem] = []
        self.allow_use_same_line = allow_use_same_line
        self.register_event_listeners()

    @contextmanager
    def work_context(self, command: str, patterns: List[Union[str, re.Pattern]]):
        item = QueryItem(command, patterns)
        try:
            self.__items.append(item)
            yield item
        finally:
            self.__items.remove(item)

    def query(
            self,
            command: str,
            patterns: List[Union[str, re.Pattern]],
            timeout: float = 3.0
    ) -> Optional[re.Match]:
        if self.__server.is_on_executor_thread():
            raise RuntimeError("Can't run query operation on executor thread")
        with self.work_context(command, patterns) as item:  # type: QueryItem
            self.__server.execute(item.command)
            return item.get(timeout=timeout)

    def on_info(self, server: "PluginServerInterface", info: "Info"):
        if info.is_from_server and not self.__items:
            for item in self.__items:
                match = None
                for p in item.patterns:
                    match = p.fullmatch(str(info.content))
                    if match is not None:
                        break
                if match is not None:
                    item.queue.put(match)
                    if not self.allow_use_same_line:
                        break

    def register_event_listeners(self):
        self.__server.register_event_listener(MCDRPluginEvents.GENERAL_INFO, self.on_info)



import dataclasses
from typing import Optional, Any, List, Union, TypeVar

from typing_extensions import final, Annotated

from pydantic import (
    BaseModel,
    field_validator,
    BeforeValidator,
    WrapSerializer,
    SerializerFunctionWrapHandler,
    ConfigDict,
    model_serializer,
    SerializationInfo
)

from calyx_lib.config.yaml import CommentCarrier, ConfigComment, CommentTextWrapper

T = TypeVar("T")


@final
class __Blank:
    @classmethod
    def ser_blank(cls, value: Any, handler: SerializerFunctionWrapHandler):
        return None if isinstance(value, cls) else handler(value)

    @classmethod
    def val_blank(cls, value: Any):
        return BLANK if value is None else value


BLANK = __Blank()
Blankable = Annotated[
    Union[T, __Blank],
    BeforeValidator(__Blank.val_blank),
    WrapSerializer(__Blank.ser_blank)
]


class PydanticValidationErrorMessage(BaseModel):
    ctx: Optional[Any] = None
    input: Any
    loc: List[Union[str, int]]
    msg: str
    type: str
    url: str

    @field_validator('loc', mode='before')
    @classmethod
    def model_val_loc(cls, v: Any):
        return list(v)


@dataclasses.dataclass
class ConfigSerializationContext:
    export_carrier: bool = True
    global_wrapper: "CommentTextWrapper" = lambda text: text


class CommentedModel(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra='ignore',
        use_enum_values=True
    )

    @model_serializer(mode='wrap')
    def ser_model(
            self, nxt, info: SerializationInfo
    ) -> Union[dict, CommentCarrier]:
        target_dict: dict = nxt(self)
        if not (
                isinstance(info.context, ConfigSerializationContext)
                and
                info.context.export_carrier
        ):
            return target_dict

        carrier = CommentCarrier(target_dict)
        carrier.global_wrapper = info.context.global_wrapper
        for k, f in self.model_fields.items():
            for meta in f.metadata:
                if isinstance(meta, ConfigComment):
                    carrier.set_comment_to_nested_key([k], meta)
        return carrier



import os

from setuptools import setup, find_packages

from calyx_lib import constants


here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'requirements.txt')) as f:
    REQUIRED = list(filter(None, map(str.strip, f)))
    print('REQUIRED = {}'.format(REQUIRED))

with open(os.path.join(here, 'README.md'), encoding='utf8') as f:
    LONG_DESCRIPTION = f.read()

CLASSIFIERS = [
    # https://pypi.org/classifiers/
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    # 'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Programming Language :: Python :: 3.13',
]

setup(
    name=constants.PACKAGE_NAME,
    version=constants.VERSION_PYPI,
    description=constants.DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url=constants.GITHUB_URL,
    author_email="ra1ny_yuki@outlook.com",
    author="Ra1ny_Yuki",
    classifiers=CLASSIFIERS,
    install_requires=REQUIRED,
    packages=find_packages(exclude=['tests', '*.tests', '*.tests.*', 'tests.*'])
)



"""
Custom logger for MCDR

Modified from https://github.com/MCDReforged/MCDReforged/
Lesser GNU Public License v3
"""
import datetime
import itertools
import logging
import os
import sys
import time
import zipfile
from contextlib import contextmanager
from pathlib import Path
from threading import local, Lock
from typing import Dict, Optional, IO, Type, Any

from colorama import Fore, Style
from colorlog import ColoredFormatter
from mcdreforged.api.types import ServerInterface

from calyx_lib.utils import (
    clean_minecraft_color_code,
    clean_console_color_code,
    touch_directory
)


class DummyLogger(logging.Logger):
    def _log(self, *args, **kwargs):
        pass


class ZippingDayRotatingFileHandler(logging.FileHandler):
    def __init__(self, file_path: str, rotate_day_count: int):
        self.rotate_day_count = rotate_day_count
        self.file_path = Path(file_path)
        self.dir_path = self.file_path.parent

        self.last_rover_date: Optional[datetime.date] = None
        self.last_record_date: Optional[datetime.date] = None
        self.try_rotate()

        super().__init__(file_path, encoding='utf8')

    def emit(self, record: logging.LogRecord) -> None:
        try:
            self.try_rotate()
            super().emit(record)
        except Exception:       # type: ignore
            self.handleError(record)

    def try_rotate(self):
        current = datetime.datetime.now().date()

        if (
            self.last_rover_date is None
            or
            (current - self.last_rover_date).days >= self.rotate_day_count
        ):
            self.do_rotate(
                self.last_record_date and self.last_record_date.strftime('%Y-%m-%d')
            )
            self.last_rover_date = current

        self.last_record_date = current

    def do_rotate(self, base_name: Optional[str] = None):
        if not self.file_path.is_file():
            return

        inited = hasattr(self, 'stream')
        if inited:
            self.stream.close()
        try:
            if base_name is None:
                try:
                    log_time = time.localtime(self.file_path.stat().st_mtime)
                except (OSError, OverflowError, ValueError):
                    log_time = time.localtime()
                base_name = time.strftime('%Y-%m-%d', log_time)
            for counter in itertools.count(start=1):
                zip_path = self.dir_path / '{}-{}.zip'.format(base_name, counter)
                if not zip_path.is_file():
                    break
            else:
                raise RuntimeError('should already able to get a valid zip path')
            with zipfile.ZipFile(zip_path, 'w') as zipf:
                zipf.write(
                    self.file_path,
                    arcname=self.file_path.name,
                    compress_type=zipfile.ZIP_DEFLATED
                )

            try:
                self.file_path.unlink()
            except OSError:
                # failed to delete the old log file
                # might due to another MCDR instance being running
                # delete the rotated zip file to avoid duplication
                try:
                    zip_path.unlink()
                except OSError:
                    pass
                raise
        finally:
            if inited:
                self.stream = self._open()


class _SyncWriteStream:
    """
    A stream wrapper with its method "write" synchronized.
    All accesses to other attributes are forwarded to the wrapped stream
    """
    def __init__(self, stream: IO[str]):
        self.sws_stream = stream
        self.sws_lock = Lock()

    def write(self, s: str):
        with self.sws_lock:
            self.sws_stream.write(s)

    def __getattribute__(self, item: str):
        if item in ('write', 'sws_stream', 'sws_lock'):
            return object.__getattribute__(self, item)
        else:
            return self.sws_stream.__getattribute__(item)


class SyncStdoutStreamHandler(logging.StreamHandler):
    __sws = _SyncWriteStream(sys.stdout)

    def __init__(self):
        super().__init__(type(self).__sws)

    @classmethod
    def update_stdout(cls, stream: IO[str]):
        cls.__sws.sws_stream = stream

    @classmethod
    def write_direct(cls, s: str):
        cls.__sws.write(s)


class MCColorFormatControl:
    MC_CODE_ITEMS: Dict[str, "Fore"] = {    # type: ignore
        '§0': Fore.BLACK,
        '§1': Fore.BLUE,
        '§2': Fore.GREEN,
        '§3': Fore.CYAN,
        '§4': Fore.RED,
        '§5': Fore.MAGENTA,
        '§6': Fore.YELLOW,
        '§7': Fore.WHITE + Style.DIM,
        '§8': Fore.WHITE + Style.DIM,
        '§9': Fore.LIGHTBLUE_EX,
        '§a': Fore.LIGHTGREEN_EX,
        '§b': Fore.LIGHTCYAN_EX,
        '§c': Fore.LIGHTRED_EX,
        '§d': Fore.LIGHTMAGENTA_EX,
        '§e': Fore.LIGHTYELLOW_EX,
        '§f': Fore.WHITE,
    }

    # global flag
    console_color_disabled = False

    __TLS = local()

    @classmethod
    @contextmanager
    def disable_minecraft_color_code_transform(cls):
        cls.__set_mc_code_trans_disable(True)
        try:
            yield
        finally:
            cls.__set_mc_code_trans_disable(False)

    @classmethod
    def __is_mc_code_trans_disabled(cls) -> bool:
        try:
            return cls.__TLS.mc_code_trans
        except AttributeError:
            cls.__set_mc_code_trans_disable(False)
            return False

    @classmethod
    def __set_mc_code_trans_disable(cls, state: bool):
        cls.__TLS.mc_code_trans = state

    @classmethod
    def modify_message_text(cls, text: str) -> str:
        if not cls.__is_mc_code_trans_disabled():
            # minecraft code -> console code
            if '§' in text:
                for mc_code, console_code in cls.MC_CODE_ITEMS.items():
                    if mc_code in text:
                        text = text.replace(mc_code, console_code)
                # clean the rest of minecraft codes
                text = clean_minecraft_color_code(text)
        if cls.console_color_disabled:
            text = clean_console_color_code(text)
        return text


class MCDReforgedFormatter(ColoredFormatter):
    def formatMessage(self, record: logging.LogRecord):
        text = super().formatMessage(record)
        return MCColorFormatControl.modify_message_text(text)


class NoColorFormatter(logging.Formatter):
    def formatMessage(self, record: logging.LogRecord):
        return clean_console_color_code(super().formatMessage(record))


class PluginIdAwareFormatter(logging.Formatter):
    PLUGIN_ID_KEY = 'plugin_id'

    def __init__(
            self, fmt_class: Type[logging.Formatter], fmt_str_with_key: str, **kwargs
    ):
        super().__init__()
        fmt_str_without_key = fmt_str_with_key.replace(' [%(plugin_id)s]', '')
        if fmt_str_without_key == fmt_str_with_key:
            raise ValueError(fmt_str_with_key)

        self.fmt_with_key = fmt_class(fmt_str_with_key, **kwargs)
        self.fmt_without_key = fmt_class(fmt_str_without_key, **kwargs)

    def format(self, record: logging.LogRecord):
        if hasattr(record, self.PLUGIN_ID_KEY):
            return self.fmt_with_key.format(record)
        else:
            return self.fmt_without_key.format(record)


class BlossomLogger(logging.Logger):
    DEFAULT_NAME = "Calyx"
    ROTATE_DAY_COUNT = 7
    LOG_COLORS = {
        'DEBUG': 'blue',
        'INFO': 'green',
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'bold_red',
    }
    SECONDARY_LOG_COLORS = {
        'message': {
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'red'
        }
    }

    FILE_FORMATTER = PluginIdAwareFormatter(
        NoColorFormatter,
        '[%(name)s] [%(asctime)s.%(msecs)d] [%(threadName)s/%(levelname)s] [%('
        'filename)s:%(lineno)d(%(funcName)s)] [%(plugin_id)s]: %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S',
    )
    CONSOLE_FORMATTER = PluginIdAwareFormatter(
        MCDReforgedFormatter,
        '[%(name)s] [%(asctime)s] [%(threadName)s/%(log_color)s%(levelname)s%(reset)s] '
        '[%(plugin_id)s]: %(message_log_color)s%(message)s%(reset)s',
        log_colors=LOG_COLORS,
        secondary_log_colors=SECONDARY_LOG_COLORS,
        datefmt='%H:%M:%S',
    )

    def __init__(self, logger_name: Optional[str] = None, plugin_id: Optional[str] = None):
        super().__init__(logger_name or self.DEFAULT_NAME)
        self.file_handler: Optional[logging.FileHandler] = None
        self.__plugin_id = plugin_id

        self.console_handler = SyncStdoutStreamHandler()
        self.console_handler.setFormatter(self.CONSOLE_FORMATTER)

        self.addHandler(self.console_handler)
        self.setLevel(logging.INFO)
        self.__verbosity: bool = False

    def set_verbosity(self, verbosity: bool):
        self.__verbosity = verbosity

    def should_log_debug(self, *args, **kwargs):
        mcdr_should_log = False
        try:
            psi = ServerInterface.psi_opt()
            if psi is not None:
                mcdr_should_log = psi.logger.should_log_debug()  # type: ignore
        except:
            pass
        return self.__verbosity or mcdr_should_log

    def _log(self, level: int, msg: Any, args: tuple, **kwargs) -> None:    # type: ignore
        if self.__plugin_id is not None:
            extra_args = kwargs.get('extra', {})
            extra_args[PluginIdAwareFormatter.PLUGIN_ID_KEY] = self.__plugin_id
            kwargs['extra'] = extra_args

        msg = str(msg)
        # noinspection PyProtectedMember
        for line in msg.splitlines():
            super()._log(level, line, args, **kwargs)

    def mdebug(self, msg: Any, *args, no_check: bool = False):
        """
        mcdr debug logging
        """
        if no_check or self.isEnabledFor(logging.DEBUG) or self.should_log_debug():
            with MCColorFormatControl.disable_minecraft_color_code_transform():
                self._log(logging.DEBUG, msg, args, stacklevel=2)

    def debug(self, msg: Any, *args, **kwargs):
        if self.isEnabledFor(logging.DEBUG) or self.should_log_debug():
            with MCColorFormatControl.disable_minecraft_color_code_transform():
                self._log(logging.DEBUG, msg, args, **kwargs, stacklevel=2)

    def set_file(self, file_path: str, no_zip_archive: bool = False):
        """

        :param file_path:
        :param no_zip_archive:
        :return:
        """
        if self.file_handler is not None:
            self.unset_file()

        if no_zip_archive:
            self.file_handler = logging.FileHandler(file_path, encoding='UTF-8')
        else:
            touch_directory(os.path.dirname(file_path))
            self.file_handler = ZippingDayRotatingFileHandler(
                file_path, self.ROTATE_DAY_COUNT
            )

        self.file_handler.setFormatter(self.FILE_FORMATTER)
        self.addHandler(self.file_handler)

    def unset_file(self):
        """
        **Not public API**

        :meta private:
        """
        if self.file_handler is not None:
            self.removeHandler(self.file_handler)
            self.file_handler.close()
            self.file_handler = None



import re
from logging import Logger
from threading import RLock
from typing import List, Optional

from mcdreforged.api.decorator import new_thread
from mcdreforged.api.event import MCDRPluginEvents
from mcdreforged.api.types import PluginServerInterface

from calyx_lib.query import CommandQueries
from calyx_lib.utils import to_camel_case


ONLINE_PLAYER_MATCH_PATTERN = [
    # <1.16
    # There are 6 of a max 100 players online: 122, abc, xxx, www, QwQ, bot_tob
    r"There are (?P<amount>[0-9]+) of a max (?P<limit>[0-9]+) players online:("
    r"?P<players>[\s\S]+)",
    # >=1.16
    # There are 2 of a max of 20 players online: Ra1ny_Yuki, xinbing
    r'There are (?P<amount>[0-9]+) of a max of (?P<limit>[0-9]+) players online:('
    r'?P<players>[\s\S]+)',
]


class OnlinePlayerRecorder:
    def __init__(
            self, server: "PluginServerInterface", logger: Optional["Logger"] = None
    ):
        self.__lock = RLock()
        self.__players: List[str] = []
        self.__server = server
        self.__logger = logger or server.logger
        self.__thread_prefix = to_camel_case(self.__server.get_self_metadata().id) + '@'
        self.__limit = 0
        self.__enabled = False
        self.__command = 'list'
        self.__queries = CommandQueries(self.__server, self.__logger)
        self.__patterns = ONLINE_PLAYER_MATCH_PATTERN

        self.register_event_listeners()

    def set_player_list_query_patterns(self, patterns: List[str]) -> None:
        """
        If the default patterns work fine, this method is not required to be called.
        Unless these patterns can't match your server return message for command list
        :param patterns: The patterns to replace the default ones
        :return: None
        """
        self.__patterns = patterns

    def set_player_list_command(self, command: str):
        self.__command = command

    def get_player_list(self, refresh: bool = False) -> List[str]:
        """
        Get player list in server
        :param refresh: If this is `true`, the values will be refreshed before return
        :return: List of player names
        """
        with self.__lock:
            if refresh:
                self.__refresh_online_players()
            return self.__players.copy()

    def get_player_limit(self, refresh: bool = False) -> int:
        """
        Get the maximum player count for this server
        :param refresh: If this is `true`, the values will be refreshed before return
        :return: Player limit value
        """
        with self.__lock:
            if refresh or self.__limit is None:
                self.__refresh_online_players()
            return self.__limit

    def __add_player(self, player: str):
        @new_thread(self.__thread_prefix + "AddOnlinePlayer")
        def __execute():
            with self.__lock:
                if self.__enabled and player not in self.__players:
                    self.__players.append(player)

        return __execute()

    def __remove_player(self, player: str):
        @new_thread(self.__thread_prefix + "RemoveOnlinePlayer")
        def __execute():
            with self.__lock:
                if self.__enabled and player in self.__players:
                    self.__players.remove(player)

        return __execute()

    def __refresh_online_players(self, timeout: int = 3):
        @new_thread(self.__thread_prefix + "RefreshOnlinePlayers")
        def __execute():
            with self.__lock:
                self.__logger.debug("Refreshing online players")
                if not self.__server.is_server_startup():
                    return

                self.__logger.debug(f"Player list command query timeout = {timeout}")
                match: Optional[re.Match] = self.__queries.query(   # type: ignore
                    self.__command, ONLINE_PLAYER_MATCH_PATTERN, timeout=timeout
                )

                if match is not None:
                    amount = match['amount']
                    self.__limit = match['limit']
                    players_string = match['players'].strip()
                    self.__players = players_string.split(', ')
                    self.__logger.debug(
                        "Player list refreshed: "
                        + ", ".join(self.__players)
                        + f" (max {self.__limit})"
                    )
                    if amount != len(self.__players):
                        self.__logger.warning(
                            "Incorrect player count found while refreshing player list"
                        )
                self.__enabled = True
        return __execute()

    def __enable_player_join(self):
        @new_thread(self.__thread_prefix + "EnablePlayerJoin")
        def __execute():
            with self.__lock:
                self.__enabled = True
                self.__logger.debug("Player list counting enabled")

        return __execute()

    def __clear_online_players(self):
        @new_thread(self.__thread_prefix + "ClearOnlinePlayers")
        def __execute():
            with self.__lock:
                self.__limit, self.__players = None, []
                self.__enabled = False
                self.__logger.debug(
                    "Cleared online player cache, player list counting disabled"
                )

        return __execute()

    def register_event_listeners(self) -> None:
        """
        If you want to make a recorder instance work in your plugin,
        This must be called in your plugin
        when plugin loaded event (on_load()) is dispatched
        :return: None
        """
        self.__queries.register_event_listeners()
        self.__server.register_event_listener(
            MCDRPluginEvents.PLUGIN_LOADED,
            lambda *args, **kwargs: self.__refresh_online_players(),
        )
        self.__server.register_event_listener(
            MCDRPluginEvents.SERVER_START,
            lambda *args, **kwargs: self.__enable_player_join(),
        )
        self.__server.register_event_listener(
            MCDRPluginEvents.PLAYER_JOINED,
            lambda _, player, __: self.__add_player(player),
        )
        self.__server.register_event_listener(
            MCDRPluginEvents.PLAYER_LEFT, lambda _, player: self.__remove_player(player)
        )
        self.__server.register_event_listener(
            MCDRPluginEvents.SERVER_STOP,
            lambda *args, **kwargs: self.__clear_online_players(),
        )